Package: shinybrms
Title: Graphical User Interface ('shiny' App) for 'brms'
Version: 1.8.1
Date: 2025-12-01
Authors@R: 
    c(person(given = "Frank",
             family = "Weber",
             role = c("aut", "cre"),
             email = "fweber144@protonmail.com",
             comment = c(ORCID = "0000-0002-4842-7922")),
      person(given = "Katja",
             family = "Ickstadt",
             role = c("ctb"),
             comment = c(ORCID = "0000-0001-5157-2496")),
      person(given = "Änne",
             family = "Glass",
             role = c("ctb"),
             comment = c(ORCID = "0000-0002-7715-9058")),
      person(given = "Thomas",
             family = "Park",
             role = c("ctb", "cph"),
             comment = "'Bootswatch' theme \"United\""),
      person(family = "Twitter, Inc.",
             role = c("ctb", "cph"),
             comment = "'Bootstrap' (basis for the 'Bootswatch' theme \"United\")"),
      person(family = "Google, LLC",
             role = c("ctb", "cph"),
             comment = "\"Open Sans\" font"))
Description: A graphical user interface (GUI) for fitting Bayesian regression
    models using the package 'brms' which in turn relies on 'Stan'
    (<https://mc-stan.org/>). The 'shinybrms' GUI is a 'shiny' app.
License: GPL-3 | file LICENSE
URL: https://fweber144.github.io/shinybrms/
BugReports: https://github.com/fweber144/shinybrms/issues
Depends: R (>= 3.6.0)
Imports: shiny (>= 1.7.0), brms (>= 2.16.0), rstan (>= 2.19.3), rlang
Suggests: cmdstanr, posterior, ggplot2, shinystan (>= 2.4.0), callr (>=
        3.4.0), rstanarm, MASS, lme4, testthat (>= 3.0.0), shinytest2
Additional_repositories: https://stan-dev.r-universe.dev/
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2025-12-01 21:58:48 UTC; frank
Author: Frank Weber [aut, cre] (ORCID: <https://orcid.org/0000-0002-4842-7922>),
  Katja Ickstadt [ctb] (ORCID: <https://orcid.org/0000-0001-5157-2496>),
  Änne Glass [ctb] (ORCID: <https://orcid.org/0000-0002-7715-9058>),
  Thomas Park [ctb, cph] ('Bootswatch' theme "United"),
  Twitter, Inc. [ctb, cph] ('Bootstrap' (basis for the 'Bootswatch' theme
    "United")),
  Google, LLC [ctb, cph] ("Open Sans" font)
Maintainer: Frank Weber <fweber144@protonmail.com>
Repository: CRAN
Date/Publication: 2025-12-01 22:30:02 UTC
Built: R 4.6.0; ; 2026-04-30 03:29:21 UTC; unix
